import logging

from vcftoolbox import (get_variant_dict, HeaderParser, get_info_dict,
                        get_vep_info, get_snpeff_info, get_vcf_handle,
                        get_variant_id, Genotype)

from puzzle.plugins import BaseVariantMixin

from puzzle.models import (Compound, Variant, Gene, Transcript,)
from puzzle.models import Genotype as puzzle_genotype
from puzzle.utils import (get_most_severe_consequence, get_omim_number,
                          get_cytoband_coord, IMPACT_SEVERITIES)

logger = logging.getLogger(__name__)

class VariantMixin(BaseVariantMixin):
    """Class to store variant specific functions for vcf plugin"""

    def variant(self, case_id, variant_id):
        """Return a specific variant.

            Args:
                case_id (str): Path to vcf file
                variant_id (str): A variant id

            Returns:
                variant (Variant): The variant object for the given id
        """
        case_obj = self.case(case_id=case_id)
        vcf_file_path = case_obj.variant_source
        head = self._get_header(vcf_file_path)

        handle = get_vcf_handle(infile=vcf_file_path)
        relevant_lines = (line for line in handle if not line.startswith('#'))
        for index, variant_line in enumerate(relevant_lines):
            index += 1
            line_id = get_variant_id(variant_line=variant_line).lstrip('chrCHR')
            if line_id == variant_id:
                return self._format_variant(
                    variant_line=variant_line,
                    index=index,
                    case_obj=case_obj,
                    head=head
                )

        return None

    def variants(self, case_id, skip=0, count=30, filters=None):
        """Return all variants in the VCF.

            Args:
                case_id (str): Path to a vcf file (for this adapter)
                skip (int): Skip first variants
                count (int): The number of variants to return
                filters (dict): A dictionary with filters. Currently this will
                look like: {
                    gene_list: [] (list of hgnc ids),
                    frequency: None (float),
                    cadd: None (float),
                    sv_len: None (float),
                    consequence: [] (list of consequences),
                    is_lof: None (Bool),
                    genetic_models [] (list of genetic models)
                    sv_type: List (list of sv types),
                }
        """
        filters = filters or {}
        case_obj = self.case(case_id=case_id)
        limit = count + skip

        genes = None
        if filters.get('genes'):
            genes = set(filters['genes'])

        frequency = None
        if filters.get('frequency'):
            frequency = float(filters['frequency'])

        cadd = None
        if filters.get('cadd'):
            cadd_score = float(filters['cadd'])

        genetic_models = None
        if filters.get('genetic_models'):
            genetic_models = set(filters['genetic_models'])

        sv_len = None
        if filters.get('sv_len'):
            sv_len = float(filters['sv_len'])

        impact_severities = None
        if filters.get('impact_severities'):
            impact_severities = set(filters['impact_severities'])

        vcf_file_path = case_obj.variant_source

        head = self._get_header(vcf_file_path)

        raw_variants = self._get_filtered_variants(vcf_file_path, filters)

        skip_index = 0
        for index, variant_line in enumerate(raw_variants):
            index += 1
            if skip_index >= skip:
                variant_obj = self._format_variant(
                     variant_line=variant_line,
                     index=index,
                     case_obj=case_obj,
                     head=head
                )

                if genes and variant_obj:
                    if not set(variant_obj['genes']).intersection(genes):
                        variant_obj = None

                if impact_severities and variant_obj:
                    if not variant_obj['impact_severity'] in impact_severities:
                        variant_obj = None

                if frequency and variant_obj:
                    if variant_obj['max_freq'] > frequency:
                        variant_obj = None

                if cadd and variant_obj:
                    if variant_obj['cadd_score'] < cadd_score:
                        variant_obj = None

                if genetic_models and variant_obj:
                    if not set(variant_obj.genetic_models).intersection(genetic_models):
                        variant_obj = None

                if sv_len and variant_obj:
                    if variant_obj.sv_len < sv_len:
                        variant_obj = None

                if variant_obj:
                    skip_index += 1

                    if skip_index <= limit:
                        yield variant_obj
                    else:
                        break
            else:
                skip_index += 1

    def _get_header(self, vcf_file_path):
        """Parse the header and return a header object

            Args:
                vcf_file_path(str): Path to vcf

            Returns:
                head: A HeaderParser object
        """
        logger.info("Parsing header of file {0}".format(vcf_file_path))
        head = HeaderParser()
        handle = get_vcf_handle(infile=vcf_file_path)
        # Parse the header
        for line in handle:
            line = line.rstrip()
            if line.startswith('#'):
                if line.startswith('##'):
                    head.parse_meta_data(line)
                else:
                    head.parse_header_line(line)
            else:
                break

        handle.close()

        return head

    def _get_filtered_variants(self, vcf_file_path, filters={}):
        """Check if variants follows the filters

            This function will try to make filters faster for the vcf adapter

            Args:
                vcf_file_path(str): Path to vcf
                filters (dict): A dictionary with filters

            Yields:
                varian_line (str): A vcf variant line
        """

        genes = set()
        consequences = set()
        sv_types = set()

        if filters.get('gene_ids'):
            genes = set([gene_id.strip() for gene_id in filters['gene_ids']])

        if filters.get('consequence'):
            consequences = set(filters['consequence'])

        if filters.get('sv_types'):
            sv_types = set(filters['sv_types'])

        logger.info("Get variants from {0}".format(vcf_file_path))

        handle = get_vcf_handle(infile=vcf_file_path)
        for variant_line in handle:
            if not variant_line.startswith('#'):
                keep_variant = True

                if genes and keep_variant:
                    keep_variant = False
                    for gene in genes:
                        if "{0}".format(gene) in variant_line:
                            keep_variant = True
                            break

                if consequences and keep_variant:
                    keep_variant = False
                    for consequence in consequences:
                        if consequence in variant_line:
                            keep_variant = True
                            break

                if sv_types and keep_variant:
                    keep_variant = False
                    for sv_type in sv_types:
                        if sv_type in variant_line:
                            keep_variant = True
                            break

                if keep_variant:
                    yield variant_line

    def _add_compounds(self, variant, info_dict):
        """Check if there are any compounds and add them to the variant

        """
        compound_entry = info_dict.get('Compounds')
        if compound_entry:
            for family_annotation in compound_entry.split(','):
                compounds = family_annotation.split(':')[-1].split('|')
                for compound in compounds:
                    splitted_compound = compound.split('>')

                    compound_score = None
                    if len(splitted_compound) > 1:
                        compound_id = splitted_compound[0]
                        compound_score = splitted_compound[-1]

                    variant.add_compound(Compound(
                        variant_id=compound_id,
                        combined_score=compound_score
                    ))

    def _get_vep_transcripts(self, transcript_info):
        """Get all transcripts for a variant

            Args:
                transcript_info (dict): A dict with vep info

            Returns:
                transcript (puzzle.models.Transcript): A Transcripts
        """
        transcript = Transcript(
                hgnc_symbol = transcript_info.get('SYMBOL'),
                transcript_id = transcript_info.get('Feature'),
                ensembl_id = transcript_info.get('Gene'),
                biotype = transcript_info.get('BIOTYPE'),
                consequence = transcript_info.get('Consequence'),
                strand = transcript_info.get('STRAND'),
                sift = transcript_info.get('SIFT'),
                polyphen = transcript_info.get('PolyPhen'),
                exon = transcript_info.get('EXON'),
                HGVSc = transcript_info.get('HGVSc'),
                HGVSp = transcript_info.get('HGVSp')
            )
        return transcript

    def _get_snpeff_transcripts(self, transcript_info):
        """Get all transcripts for a variant

            Args:
                transcript_info (dict): A dict with snpeff info

            Returns:
                transcript (puzzle.models.Transcript): A Transcripts
        """
        transcript = Transcript(
                hgnc_symbol = transcript_info.get('Gene_Name'),
                transcript_id = transcript_info.get('Feature'),
                ensembl_id = transcript_info.get('Gene_ID'),
                biotype = transcript_info.get('Transcript_BioType'),
                consequence = transcript_info.get('Annotation'),
                exon = transcript_info.get('Rank'),
                HGVSc = transcript_info.get('HGVS.c'),
                HGVSp = transcript_info.get('HGVS.p')
            )
        return transcript

    def _format_variant(self, variant_line, index, case_obj, head):
        """Return variant objects

            Args:
                raw_variants (Iterable): An iterable with variant lines
                case_obj (puzzle.nodels.Case): A case object

        """
        header_line = head.header

        # Get the individual ids for individuals in vcf file
        vcf_individuals = set([ind_id for ind_id in head.individuals])

        variant_columns = ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER']

        vep_header = head.vep_columns
        snpeff_header = head.snpeff_columns

        #Create a variant dict:
        variant_dict =  get_variant_dict(
            variant_line = variant_line,
            header_line = header_line
        )
        variant_dict['CHROM'] = variant_dict['CHROM'].lstrip('chrCHR')
        #Crreate a info dict:
        info_dict = get_info_dict(
            info_line = variant_dict['INFO']
        )
        #Check if vep annotation:
        vep_string = info_dict.get('CSQ')

        #Check if snpeff annotation:
        snpeff_string = info_dict.get('ANN')

        if vep_string:
            #Get the vep annotations
            vep_info = get_vep_info(
                vep_string = vep_string,
                vep_header = vep_header
            )

        elif snpeff_string:
            #Get the vep annotations
            snpeff_info = get_snpeff_info(
                snpeff_string = snpeff_string,
                snpeff_header = snpeff_header
            )

        variant = Variant(
            **{column: variant_dict.get(column, '.')
                for column in variant_columns}
            )

        logger.debug("Creating a variant object of variant {0}".format(
            variant.get('variant_id')))

        variant['index'] = index
        logger.debug("Updating index to: {0}".format(
            index))

        variant['start'] = int(variant_dict['POS'])


        if self.variant_type == 'sv':
            other_chrom = variant['CHROM']
            # If we have a translocation:
            if ':' in variant_dict['ALT'] and not '<' in variant_dict['ALT']:
                other_coordinates = variant_dict['ALT'].strip('ACGTN[]').split(':')
                other_chrom = other_coordinates[0].lstrip('chrCHR')
                other_position = other_coordinates[1]
                variant['stop'] = other_position

                #Set 'infinity' to length if translocation
                variant['sv_len'] = float('inf')
            else:
                variant['stop'] = int(info_dict.get('END', variant_dict['POS']))
                variant['sv_len'] = variant['stop'] - variant['start']

            variant['stop_chrom'] = other_chrom

        else:
            variant['stop'] = int(variant_dict['POS']) + \
                (len(variant_dict['REF']) - len(variant_dict['ALT']))

        variant['sv_type'] = info_dict.get('SVTYPE')
        variant['cytoband_start'] = get_cytoband_coord(
                                        chrom=variant['CHROM'],
                                        pos=variant['start'])
        if variant.get('stop_chrom'):
            variant['cytoband_stop'] = get_cytoband_coord(
                                        chrom=variant['stop_chrom'],
                                        pos=variant['stop'])

        # It would be easy to update these keys...
        thousand_g = info_dict.get('1000GAF')
        if thousand_g:
            logger.debug("Updating thousand_g to: {0}".format(
                thousand_g))
            variant['thousand_g'] = float(thousand_g)
            variant.add_frequency('1000GAF', variant.get('thousand_g'))

        #SV specific tag for number of occurances
        occurances = info_dict.get('OCC')
        if occurances:
            logger.debug("Updating occurances to: {0}".format(
                occurances))
            variant['occurances'] = float(occurances)
            variant.add_frequency('OCC', occurances)

        cadd_score = info_dict.get('CADD')
        if cadd_score:
            logger.debug("Updating cadd_score to: {0}".format(
                cadd_score))
            variant['cadd_score'] = float(cadd_score)

        rank_score_entry = info_dict.get('RankScore')
        if rank_score_entry:
            for family_annotation in rank_score_entry.split(','):
                rank_score = family_annotation.split(':')[-1]
            logger.debug("Updating rank_score to: {0}".format(
                rank_score))
            variant['rank_score'] = float(rank_score)

        genetic_models_entry = info_dict.get('GeneticModels')
        if genetic_models_entry:
            genetic_models = []
            for family_annotation in genetic_models_entry.split(','):
                for genetic_model in family_annotation.split(':')[-1].split('|'):
                    genetic_models.append(genetic_model)
            logger.debug("Updating rank_score to: {0}".format(
                rank_score))
            variant['genetic_models'] = genetic_models

        #Add genotype calls:
        for individual in case_obj.individuals:
            sample_id = individual.ind_id

            if sample_id in vcf_individuals:

                raw_call = dict(zip(
                    variant_dict['FORMAT'].split(':'),
                    variant_dict[sample_id].split(':'))
                )

                genotype = Genotype(**raw_call)

                variant.add_individual(puzzle_genotype(
                    sample_id = sample_id,
                    genotype = genotype.genotype,
                    case_id = individual.case_name,
                    phenotype = individual.phenotype,
                    ref_depth = genotype.ref_depth,
                    alt_depth = genotype.alt_depth,
                    genotype_quality = genotype.genotype_quality,
                    depth = genotype.depth_of_coverage,
                    supporting_evidence = genotype.supporting_evidence,
                    pe_support = genotype.pe_support,
                    sr_support = genotype.sr_support,
                ))

        # Add transcript information:
        gmaf = None
        if vep_string:
            for transcript_info in vep_info:
                transcript = self._get_vep_transcripts(transcript_info)
                gmaf_raw = transcript_info.get('GMAF')
                if gmaf_raw:
                    gmaf = float(gmaf_raw.split(':')[-1])
                variant.add_transcript(transcript)

        if gmaf:
            variant.add_frequency('GMAF', gmaf)
            if not variant.thousand_g:
                variant.thousand_g = gmaf

        elif snpeff_string:
            for transcript_info in snpeff_info:
                transcript = self._get_snpeff_transcripts(transcript_info)
                variant.add_transcript(transcript)

        most_severe_consequence = get_most_severe_consequence(
            variant['transcripts']
        )
        if most_severe_consequence:
            variant['most_severe_consequence'] = most_severe_consequence

            variant['impact_severity'] = IMPACT_SEVERITIES.get(most_severe_consequence)

        for gene in self._get_genes(variant):
            variant.add_gene(gene)

        self._add_compounds(variant=variant, info_dict=info_dict)

        return variant




##TODO mock a gemini db
import pytest
from puzzle.plugins import GeminiPlugin
from puzzle.models import Individual, Variant, DotDict

def test_get_all_variants(gemini_path):
    """Test to get some variants from the gemini adapter"""
    adapter = GeminiPlugin(db=gemini_path)
    variants = []
    for variant in adapter.variants('643594'):
        variants.append(variant)

    assert len(variants) == 14

def test_get_variants(gemini_path):
    """Test to get some variants from the gemini adapter"""
    adapter = GeminiPlugin(db=gemini_path)
    variants = []
    for variant in adapter.variants('643594', count=5):
        variants.append(variant)

    assert len(variants) == 5

def test_variant(gemini_path):
    """Test to get one variant"""
    adapter = GeminiPlugin(db=gemini_path)
    variant = adapter.variant(
        case_id='643594',
        variant_id=4
    )

    assert variant['CHROM'] == '6'
    assert variant['POS'] == '32487163'
    assert type(variant['genes']) == type([])

def test_is_variant(case_obj):
    adapter = GeminiPlugin()
    genotypes = ['G/A', 'G/A', 'G/G', 'G/G', './.']
    ind_objs = [
        Individual(ind_id=0, ind_index=0),
        Individual(ind_id=1, ind_index=1),
    ]
    #Mock a gemini variant
    gemini_variant = {
        'gts':genotypes,
        'alt':'A'
    }
    
    assert adapter._is_variant(gemini_variant, ind_objs)
    
    #Check with individuals that are hom ref
    ind_objs = [
        Individual(ind_id=0, ind_index=2),
        Individual(ind_id=1, ind_index=3),
    ]

    assert not adapter._is_variant(gemini_variant, ind_objs)

    #Check with individual that has no call
    ind_objs = [
        Individual(ind_id=0, ind_index=4),
    ]

    assert not adapter._is_variant(gemini_variant, ind_objs)

def test_build_gemini_query():
    adapter = GeminiPlugin()
    query = "SELECT * from variants"
    extra_info = "max_aaf_all < 0.01"
    new_query = adapter.build_gemini_query(query, extra_info)
    assert new_query == "SELECT * from variants WHERE max_aaf_all < 0.01"
    
    extra_info = "cadd_score > 10"
    new_query = adapter.build_gemini_query(new_query, extra_info)
    assert new_query == "SELECT * from variants WHERE max_aaf_all < 0.01 AND cadd_score > 10"

def test_get_genotypes(gemini_variant):
    adapter = GeminiPlugin()
    ind = DotDict()
    ind.ind_index = 0
    ind.ind_id = '1'
    ind.case_id = 'Case_1'
    ind.phenotype = 2
    ind_objs = [ind]
    
    individual = adapter._get_genotypes(gemini_variant, ind_objs)[0]
    
    assert individual.sample_id == ind.ind_id
    assert individual.sample_id == ind.ind_id
    assert individual.genotype == 'G/A'
    assert individual.case_id == ind.case_id
    assert individual.phenotype == ind.phenotype
    assert individual.ref_depth == 10
    assert individual.alt_depth == 7
    assert individual.depth == 17
    assert individual.genotype_quality == 99

def test_get_transcripts(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    gemini_variant = {'variant_id': 1}
    transcripts = []
    for transcript in adapter._get_transcripts(gemini_variant):
        transcripts.append(transcript)
    
    assert len(transcripts) == 2
    
    first_transcript = transcripts[0]
    
    assert first_transcript.transcript_id == 'ENST00000370383'
    assert first_transcript.hgnc_symbol == 'EPHX4'
    assert first_transcript.consequence == 'missense_variant'
    assert first_transcript.ensembl_id == None
    assert first_transcript.biotype == 'protein_coding'
    assert first_transcript.sift == 'deleterious'
    assert first_transcript.polyphen == 'probably_damaging'

class TestFilters:

    def test_filters_frequency(self, gemini_path):
        adapter = GeminiPlugin(db=gemini_path)
        filters = {'frequency':'0.01'}
        variants = []
        for variant in adapter.variants('643594', filters=filters):
            variants.append(variant)
        assert len(variants) == 13

    def test_filters_cadd(self, gemini_path):
        adapter = GeminiPlugin(db=gemini_path)
        filters = {'cadd':'20'}
        variants = []
        for variant in adapter.variants('643594', filters=filters):
            variants.append(variant)
        assert len(variants) == 4
    
    def test_filters_impact_severities(self, gemini_path):
        adapter = GeminiPlugin(db=gemini_path)
        filters = {'impact_severities':['HIGH']}
        variants = []
        for variant in adapter.variants('643594', filters=filters):
            variants.append(variant)
        assert len(variants) == 2

    def test_filters_gene_ids(self, gemini_path):
        adapter = GeminiPlugin(db=gemini_path)
        filters = {'gene_ids':['HLA-DRB5']}
        variants = []
        for variant in adapter.variants('643594', filters=filters):
            variants.append(variant)
        assert len(variants) == 5

    def test_filters_consequence(self, gemini_path):
        adapter = GeminiPlugin(db=gemini_path)
        filters = {'consequence':['stop_gained']}
        variants = []
        for variant in adapter.variants('643594', filters=filters):
            variants.append(variant)
        assert len(variants) == 2
    


# -*- coding: utf-8 -*-
import os
import webbrowser

import click
import logging

from . import (base, family_file, family_type, version, root, mode, 
               variant_type, phenomizer)

from puzzle.plugins import SqlStore, VcfPlugin
try:
    from puzzle.plugins import GeminiPlugin
except ImportError:
    pass

from sqlite3 import DatabaseError

from puzzle.server import create_app
from puzzle.server.settings import BaseConfig

logger = logging.getLogger(__name__)


@base.command()
@click.argument('variant-source', type=click.Path(exists=True), required=False)
@click.option('--host', default='0.0.0.0', show_default=True)
@click.option('--port', default=5000, show_default=True)
@click.option('--debug', is_flag=True)
@click.option('-p', '--pattern', default='*.vcf', show_default=True)
@click.option('--no-browser', is_flag=True, help='Prevent auto-opening browser')
@phenomizer
@family_file
@family_type
@version
@root
@mode
@variant_type
@click.pass_context
def view(ctx, host, port, debug, pattern, family_file, family_type,
         variant_source, variant_type, root, mode, no_browser, phenomizer):
    """Visualize DNA variant resources.

    1. Look for variant source(s) to visualize and inst. the right plugin
    """
    root = root or ctx.obj.get('root') or os.path.expanduser("~/.puzzle")

    if os.path.isfile(root):
        logger.error("'root' can't be a file")
        ctx.abort()

    logger.info("Root directory is: {}".format(root))

    db_path = os.path.join(root, 'puzzle_db.sqlite3')
    logger.info("db path is: {}".format(db_path))

    if variant_source is None:
        if not os.path.exists(db_path):
            logger.warn("database not initialized, run 'puzzle init'")
            ctx.abort()
        phenomizer_auth = phenomizer or ctx.obj.get('phenomizer_auth')
        plugin = SqlStore(db_path, phenomizer_auth=phenomizer_auth)
        BaseConfig.STORE_ENABLED = True

    elif mode == 'vcf':
        logger.info("Initialzing VCF plugin")
        try:
            plugin = VcfPlugin(
                root_path=variant_source,
                case_lines=family_file,
                case_type=family_type,
                pattern=pattern,
                vtype=variant_type
            )
        except SyntaxError as e:
            logger.error(e.message)
            ctx.abort()

    elif mode == 'gemini':
        logger.info("Initialzing GEMINI plugin")
        try:
            plugin = GeminiPlugin(db=variant_source, vtype=variant_type)
        except NameError:
            logger.error("Need to have gemini installed to use gemini plugin")
            ctx.abort()
        except DatabaseError as e:
            logger.error("{0} is not a valid gemini db".format(variant_source))
            logger.info("variant-source has to point to a gemini databse")
            ctx.abort()

    logger.debug("Plugin setup was succesfull")

    BaseConfig.PUZZLE_BACKEND = plugin
    BaseConfig.UPLOAD_DIR = os.path.join(root, 'resources')

    app = create_app(config_obj=BaseConfig)

    if not (no_browser or debug):
        webbrowser.open_new_tab("http://{}:{}".format(host, port))

    app.run(host=host, port=port, debug=debug)



import os

import logging
import click

from . import (base, root, family_file, family_type, variant_type, mode)

from puzzle.plugins import SqlStore, VcfPlugin

try:
    from puzzle.plugins import GeminiPlugin
except ImportError:
    pass

from sqlite3 import DatabaseError

logger = logging.getLogger(__name__)

@base.command()
@click.argument('variant-source',
    type=click.Path(exists=True),
)
@family_file
@family_type
@root
@mode
@variant_type
@click.pass_context
def load(ctx, variant_source, family_file, family_type, root, mode,
        variant_type):
    """
    Load a variant source into the database.

    If no database was found run puzzle init first.
    
    1. VCF: If a vcf file is used it can be loaded with a ped file
    2. GEMINI: Ped information will be retreived from the gemini db
    """
    root = root or ctx.obj.get('root') or os.path.expanduser("~/.puzzle")
    
    if os.path.isfile(root):
        logger.error("'root' can't be a file")
        ctx.abort()

    logger.info("Root directory is: {}".format(root))

    db_path = os.path.join(root, 'puzzle_db.sqlite3')
    logger.info("db path is: {}".format(db_path))

    if not os.path.exists(db_path):
        logger.warn("database not initialized, run 'puzzle init'")
        ctx.abort()

    logger.debug('Set puzzle backend to {0}'.format(mode))

    logger.debug('Set variant type to {0}'.format(variant_type))

    if mode == 'vcf':
        logger.info("Initialzing VCF plugin")

        try:
            plugin = VcfPlugin(
                root_path=variant_source,
                case_lines=family_file,
                case_type=family_type,
                vtype=variant_type
            )
        except SyntaxError as e:
            logger.error(e.message)
            ctx.abort()

    elif mode == 'gemini':
        logger.debug("Initialzing GEMINI plugin")
        try:
            plugin = GeminiPlugin(db=variant_source, vtype=variant_type)
        except NameError:
            logger.error("Need to have gemini installed to use gemini plugin")
            ctx.abort()
        except DatabaseError as e:
            logger.error("{} is not a valid gemini db".format(variant_source))
            logger.info("variant-source has to point to a gemini database")
            ctx.abort()

    logger.debug("Plugin setup was succesfull")
    # from gemini can create multiple cases
    store = SqlStore(db_path)

    for case_obj in plugin.cases():
        if store.case(case_id=case_obj.case_id).case_id == case_obj.case_id:
            logger.warn("{} already exists in the database"
                        .format(case_obj.case_id))
            continue

        # extract case information
        logger.debug("adding case: {}".format(case_obj.case_id))
        store.add_case(case_obj, vtype=variant_type, mode=mode)



# -*- coding: utf-8 -*-
from datetime import datetime

from sqlalchemy import Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

from puzzle.models.mixins import PedigreeHumanMixin

# base for declaring a mapping
BASE = declarative_base()


class Case(BASE):
    """
    This is the class for storing cases in the database.

    A case will be related to Individuals
    """
    __tablename__ = "case"
    id = Column(Integer, primary_key=True)
    case_id = Column(String(32), unique=True)
    name = Column(String(32)) # This is the display name
    variant_source = Column(String)
    variant_type = Column(String) # snv or sv
    variant_mode = Column(String) # vcf or gemini
    pedigree = Column(String) # For storing madeline info
    created_at = Column(DateTime, default=datetime.now)
    updated_at = Column(DateTime, default=datetime.now, onupdate=datetime.now)

    @property
    def resources(self):
        """Return resources for all individuals."""
        for individual in self.individuals:
            for resource in individual.resources:
                yield resource

    @property
    def phenotypes(self):
        """Return all phenotypes for included individuals."""
        for individual in self.individuals:
            for term in individual.phenotypes:
                yield term

    def phenotype_ids(self):
        """Return only phenotype ids for included individuals."""
        return [term.phenotype_id for term in self.phenotypes]

    def __repr__(self):
        return("<Case(id:{self.id}, case_id:{self.case_id}, name:{self.name},"\
                " variant_source:{self.variant_source}, variant_type:"\
                "{self.variant_type}, variant_mode:{self.variant_mode})>".format(self=self))


class Individual(BASE, PedigreeHumanMixin):
    """
    This is the class for storing individuals in the database.

    A Individual belongs to a case
    """
    __tablename__ = "individual"
    id = Column(Integer, primary_key=True)
    ind_id = Column(String(32))
    name = Column(String(32))
    mother = Column(String(32))
    father = Column(String(32))
    sex = Column(String(32))
    phenotype = Column(String(32))
    ind_index = Column(Integer)
    variant_source = Column(String(32))
    bam_path = Column(String(32))
    case_id = Column(Integer, ForeignKey("case.id"))
    case = relationship(Case, backref=("individuals"))

    @property
    def case_name(self):
        """Fetch display name of case."""
        return self.case.name

    def __repr__(self):
        return("<Individual(id:{self.id}, ind_id:{self.ind_id}, mother:{self.mother},"\
                " father{self.father}, sex:{self.sex}, phenotype.{self.phenotype}"\
                ", ind_index:{self.ind_index}, variant_source:"\
                "{self.variant_source}, case_id:{self.case_id})>".format(self=self))



import sys
import logging

from pprint import pprint as pp
from datetime import datetime
import click

from gemini import GeminiQuery


def check_genotypes(gemini_variant, index):
    """Check the genotypes for a variant"""
    print(gemini_variant['gts'][index])
    print(gemini_variant['gt_types'][index])

@click.command()
@click.argument('db',
    type=click.Path(exists=True),
    # type=click.Path(exists=True),
)
@click.option('-v', '--variant',
    type=int
)
@click.option('--value',
)
@click.option('--query',
    default="SELECT * from variants"
)
def test_gemini(db, variant, value, query):
    """Test a gemini db"""
    gq = GeminiQuery(db)
    
    # query = "SELECT * from variant_impacts WHERE variant_id = 1"
    # query = "SELECT * from samples"
    
    # query = "gemini comp_hets"
    
    # if variant:
    #     query = "SELECT * from variants WHERE (aaf_1kg_all < {0} or aaf_1kg_all is Null)".format(variant)
    # else:
    #      query = "select * from variants"
    
    # query = "SELECT * from variants WHERE impact_severity in ('HIGH', 'LOW')"
    gq.run(query)
    sample2idx = gq.sample_to_idx
    
    for index, variant in enumerate(gq):
        for key in variant.keys():
            if value:
                if key == value:
                    print("{0}: {1}".format(key, variant[key]))
            else:
                print("'{0}': {1},".format(key, variant[key]))
        print("")
        # check_genotypes(variant, 0)
        # check_genotypes(variant, 1)
        # check_genotypes(variant, 2)
    
    print(sample2idx)
    print(index)


if __name__ == '__main__':
    test_gemini()



# -*- coding: utf-8 -*-
"""
puzzle.plugins.sql.store
~~~~~~~~~~~~~~~~~~
"""
import itertools
import logging
import os

import phizz
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from puzzle.models import Case as BaseCase
from puzzle.models import Individual as BaseIndividual
from puzzle.models.sql import (BASE, Case, Individual, PhenotypeTerm, GeneList,
                               CaseGenelistLink, Resource, GeminiQuery)
from puzzle.plugins import VcfPlugin, Plugin

try:
    from puzzle.plugins import GeminiPlugin
except ImportError as e:
    pass
from puzzle.utils import hpo_genes

logger = logging.getLogger(__name__)


class Store(Plugin):

    """SQLAlchemy-based database object.

    .. note::
        For testing pourposes use ``:memory:`` as the ``path`` argument to
        set up in-memory (temporary) database.

    Args:
        uri (Optional[str]): path/URI to the database to connect to
        debug (Optional[bool]): whether to output logging information
        phenomizer_auth (Optional[tuple]): username and password

    Attributes:
        uri (str): path/URI to the database to connect to
        engine (class): SQLAlchemy engine, defines what database to use
        session (class): SQLAlchemy ORM session, manages persistance
        query (method): SQLAlchemy ORM query builder method
        classes (dict): bound ORM classes
    """

    def __init__(self, uri=None, debug=False, vtype='snv',
                 phenomizer_auth=None):
        super(Store, self).__init__()
        self.uri = uri
        if uri:
            self.connect(uri, debug=debug)
        self.variant_type = vtype
        self.phenomizer_auth = phenomizer_auth

        # ORM class shortcuts to enable fetching models dynamically
        # self.classes = {'gene': Gene, 'transcript': Transcript,
        #                 'exon': Exon, 'sample': Sample}

    def init_app(self, app):
        pass

    def connect(self, db_uri, debug=False):
        """Configure connection to a SQL database.

        Args:
            db_uri (str): path/URI to the database to connect to
            debug (Optional[bool]): whether to output logging information
        """
        kwargs = {'echo': debug, 'convert_unicode': True}
        # connect to the SQL database
        if 'mysql' in db_uri:
            kwargs['pool_recycle'] = 3600
        elif '://' not in db_uri:
            logger.debug("detected sqlite path URI: {}".format(db_uri))
            db_path = os.path.abspath(os.path.expanduser(db_uri))
            db_uri = "sqlite:///{}".format(db_path)

        self.engine = create_engine(db_uri, **kwargs)
        logger.debug('connection established successfully')
        # make sure the same engine is propagated to the BASE classes
        BASE.metadata.bind = self.engine
        # start a session
        self.session = scoped_session(sessionmaker(bind=self.engine))
        # shortcut to query method
        self.query = self.session.query
        return self

    @property
    def dialect(self):
        """Return database dialect name used for the current connection.
        Dynamic attribute.
        Returns:
            str: name of dialect used for database connection
        """
        return self.engine.dialect.name

    def set_up(self, reset=False):
        """Initialize a new database with the default tables and columns.
        Returns:
            Store: self
        """
        if reset:
            self.tear_down()

        logger.info("Creating database")
        # create the tables
        BASE.metadata.create_all(self.engine)
        return self

    def tear_down(self):
        """Tear down a database (tables and columns).
        Returns:
            Store: self
        """
        # drop/delete the tables
        logger.info('resetting database...')
        BASE.metadata.drop_all(self.engine)
        return self

    def save(self):
        """Manually persist changes made to various elements. Chainable.

        Returns:
            Store: ``self`` for chainability
        """
        # commit/persist dirty changes to the database
        self.session.flush()
        self.session.commit()
        return self

    def add_case(self, case_obj, vtype='snv', mode='vcf', ped_svg=None):
        """Load a case with individuals.

        Args:
            case_obj (puzzle.models.Case): initialized case model
        """
        new_case = Case(case_id=case_obj.case_id,
                        name=case_obj.name,
                        variant_source=case_obj.variant_source,
                        variant_type=vtype,
                        variant_mode=mode,
                        pedigree=ped_svg)

        # build individuals
        inds = [Individual(
            ind_id=ind.ind_id,
            name=ind.name,
            mother=ind.mother,
            father=ind.father,
            sex=ind.sex,
            phenotype=ind.phenotype,
            ind_index=ind.ind_index,
            variant_source=ind.variant_source,
            bam_path=ind.bam_path,
        ) for ind in case_obj.individuals]

        new_case.individuals = inds
        self.session.add(new_case)
        self.save()
        return new_case

    def delete_case(self, case_obj):
        """Delete a case from the database

        Args:
            case_obj (puzzle.models.Case): initialized case model
        """
        for ind_obj in case_obj.individuals:
            self.delete_individual(ind_obj)
        logger.info("Deleting case {0} from database".format(case_obj.case_id))
        self.session.delete(case_obj)
        self.save()
        return case_obj

    def delete_individual(self, ind_obj):
        """Delete a case from the database

        Args:
            ind_obj (puzzle.models.Individual): initialized individual model
        """
        logger.info("Deleting individual {0} from database".format(ind_obj.ind_id))
        self.session.delete(ind_obj)
        self.save()
        return ind_obj

    def case(self, case_id):
        """Fetch a case from the database."""
        case_obj = self.query(Case).filter_by(case_id=case_id).first()
        if case_obj is None:
            case_obj = BaseCase(case_id='unknown')
        return case_obj

    def individual(self, ind_id):
        """Fetch a case from the database."""
        ind_obj = self.query(Individual).filter_by(ind_id=ind_id).first()
        if ind_obj is None:
            ind_obj = BaseIndividual(ind_id='unknown')
        return ind_obj

    def cases(self):
        """Fetch all cases from the database."""
        return self.query(Case)

    def get_individuals(self, ind_ids=None):
        """Fetch all individuals from the database."""
        query = self.query(Individual)
        if ind_ids:
            query = query.filter(Individual.ind_id.in_(ind_ids))
        return query

    def variants(self, case_id, skip=0, count=30, filters=None):
        """Fetch variants for a case."""
        filters = filters or {}
        logger.debug("Fetching case with case_id:{0}".format(case_id))
        case_obj = self.case(case_id)
        plugin, case_id = self.select_plugin(case_obj)
        self.filters = plugin.filters

        gene_lists = (self.gene_list(list_id) for list_id
                      in filters.get('gene_lists', []))
        nested_geneids = (gene_list.gene_ids for gene_list in gene_lists)
        gene_ids = set(itertools.chain.from_iterable(nested_geneids))

        if filters.get('gene_ids'):
            filters['gene_ids'].extend(gene_ids)
        else:
            filters['gene_ids'] = gene_ids
        variants = plugin.variants(case_id, skip, count, filters)
        return variants

    def variant(self, case_id, variant_id):
        """Fetch a single variant from variant source."""
        case_obj = self.case(case_id)
        plugin, case_id = self.select_plugin(case_obj)
        variant = plugin.variant(case_id, variant_id)
        return variant

    def add_phenotype(self, ind_obj, phenotype_id):
        """Add a phenotype term to the case."""
        if phenotype_id.startswith('HP:') or len(phenotype_id) == 7:
            logger.debug('querying on HPO term')
            hpo_results = phizz.query_hpo([phenotype_id])
        else:
            logger.debug('querying on OMIM term')
            hpo_results = phizz.query_disease([phenotype_id])

        added_terms = []
        existing_ids = set(term.phenotype_id for term in ind_obj.phenotypes)
        for result in hpo_results:
            if result['hpo_term'] not in existing_ids:
                term = PhenotypeTerm(phenotype_id=result['hpo_term'],
                                     description=result['description'])
                logger.info('adding new HPO term: %s', term.phenotype_id)
                ind_obj.phenotypes.append(term)
                added_terms.append(term)

        logger.debug('storing new HPO terms')
        self.save()

        if len(added_terms) > 0:
            self.update_hpolist(ind_obj.ind_id)

        return added_terms

    def update_hpolist(self, ind_id):
        """Update the HPO gene list for a case based on current terms."""
        ind_obj = self.individual(ind_id)
        # update the HPO gene list for the case
        hpo_list = self.case_genelist(ind_obj.case)
        hpo_results = hpo_genes(ind_obj.case.phenotype_ids(),
                                *self.phenomizer_auth)

        if hpo_results is None:
            pass
            # Why raise here?
            # raise RuntimeError("couldn't link to genes, try again")
        else:
            gene_ids = [result['gene_id'] for result in hpo_results
                        if result['gene_id']]
            hpo_list.gene_ids = gene_ids
            self.save()

    def remove_phenotype(self, ind_obj, phenotypes=None):
        """Remove multiple phenotypes from an individual."""
        if phenotypes is None:
            logger.info("delete all phenotypes related to %s", ind_obj.ind_id)
            self.query(PhenotypeTerm).filter_by(ind_id=ind_obj.id).delete()
        else:
            for term in ind_obj.phenotypes:
                if term.phenotype_id in phenotypes:
                    logger.info("delete phenotype: %s from %s",
                                term.phenotype_id, ind_obj.ind_id)
                    self.session.delete(term)
        logger.debug('persist removals')
        self.save()
        self.update_hpolist(ind_obj.ind_id)

    def gene_list(self, list_id):
        """Get a gene list from the database."""
        return self.query(GeneList).filter_by(list_id=list_id).first()

    def gene_lists(self):
        """Return all gene lists from the database."""
        return self.query(GeneList)

    def add_genelist(self, list_id, gene_ids, case_obj=None):
        """Create a new gene list and optionally link to cases."""
        new_genelist = GeneList(list_id=list_id)
        new_genelist.gene_ids = gene_ids
        if case_obj:
            new_genelist.cases.append(case_obj)

        self.session.add(new_genelist)
        self.save()
        return new_genelist

    def remove_genelist(self, list_id, case_obj=None):
        """Remove a gene list and links to cases."""
        gene_list = self.gene_list(list_id)

        if case_obj:
            # remove a single link between case and gene list
            case_ids = [case_obj.id]
        else:
            # remove all links and the list itself
            case_ids = [case.id for case in gene_list.cases]
            self.session.delete(gene_list)

        case_links = self.query(CaseGenelistLink).filter(
            CaseGenelistLink.case_id.in_(case_ids),
            CaseGenelistLink.genelist_id == gene_list.id
        )
        for case_link in case_links:
            self.session.delete(case_link)

        self.save()

    def case_genelist(self, case_obj):
        """Get or create a new case specific gene list record."""
        list_id = "{}-HPO".format(case_obj.case_id)
        gene_list = self.gene_list(list_id)

        if gene_list is None:
            gene_list = GeneList(list_id=list_id)
            case_obj.gene_lists.append(gene_list)
            self.session.add(gene_list)

        return gene_list

    def select_plugin(self, case_obj):
        """Select and initialize the correct plugin for the case."""
        if case_obj.variant_mode == 'vcf':
            logger.debug("Using vcf plugin")
            plugin = VcfPlugin(vtype=case_obj.variant_type)
        elif case_obj.variant_mode == 'gemini':
            logger.debug("Using gemini plugin")
            plugin = GeminiPlugin(vtype=case_obj.variant_type)
            plugin.db = case_obj.variant_source

        plugin.case_objs = [case_obj]

        self.variant_type = case_obj.variant_type

        case_id = case_obj.case_id
        return plugin, case_id

    def add_resource(self, name, file_path, ind_obj):
        """Link a resource to an individual."""
        new_resource = Resource(name=name, individual=ind_obj, path=file_path)
        self.session.add(new_resource)
        self.save()
        return new_resource

    def resource(self, resource_id):
        """Fetch a resource."""
        return self.query(Resource).get(resource_id)

    def delete_resource(self, resource_id):
        """Link a resource to an individual."""
        resource_obj = self.resource(resource_id)
        logger.debug("Deleting resource {0}".format(resource_obj.name))
        self.session.delete(resource_obj)
        self.save()
    
    def gemini_query(self, name):
        """Return a gemini query
        
        Args:
            name (str)
        """
        logger.debug("Looking for query with name {0}".format(name))
        return self.query(GeminiQuery).get(name=name)

    def gemini_queries(self):
        """Return all gemini queries
        
        """
        return self.query(GeminiQuery)

    def add_gemini_query(self, name, query):
        """Add a user defined gemini query
        
        Args:
            name (str)
            query (str)
        """
        logger.info("Adding query {0} with text {1}".format(name, query))
        new_query = GeminiQuery(name=name, query=query)
        self.session.add(new_query)
        self.save()
    
    def delete_gemini_query(self, name):
        """Delete a gemini query
        
        Args:
            name (str)
        """
        query_obj = self.gemini_query(name)
        logger.debug("Delete query with name {0}".format(name))
        self.session.delete(query_obj)
        self.save()



# -*- coding: utf-8 -*-
import logging
import os

from ped_parser import FamilyParser
from vcftoolbox import HeaderParser, get_vcf_handle

from puzzle.models import Case, Individual

logger = logging.getLogger(__name__)


def get_case(variant_source, case_lines=None, case_type='ped', variant_type='snv',
            variant_mode='vcf'):
        """Create a cases and populate it with individuals

            Args:
                variant_source (str): Path to vcf files
                case_lines (Iterable): Ped like lines
                case_type (str): Format of case lines

            Returns:
                case_objs (list): List with Case objects
        """
        individuals = get_individuals(
            vcf=variant_source,
            case_lines=case_lines,
            case_type=case_type,
        )
        
        #If no individuals we still need to have a case id
        case_id = os.path.basename(variant_source)
        
        for individual in individuals:
            case_id = individual.case_id

        case = Case(case_id=case_id, variant_source=variant_source,
                    name=case_id, variant_type=variant_type, 
                    variant_mode=variant_mode,
                    )

        logger.debug("Found case with case_id: {0} and name: {1}".format(
            case.case_id, case.name))

        for individual in individuals:
            case.add_individual(individual)

        return case


def get_individuals(vcf=None, case_lines=None, case_type='ped'):
        """Get the individuals from a vcf file, and/or a ped file.

            Args:
                vcf (str): Path to a vcf
                case_lines(Iterable): Ped like lines
                case_type(str): Format of ped lines

            Returns:
                individuals (generator): generator with Individuals
        """
        individuals = []

        if case_lines:
            # read individuals from ped file
            family_parser = FamilyParser(case_lines, family_type=case_type)
            families = family_parser.families
            logger.info("Found families {0}".format(
                            ','.join(list(families.keys()))))
            if len(families) != 1:
                logger.error("Only one family can be used with vcf adapter")
                raise IOError

            case_id = list(families.keys())[0]
            logger.info("Family used in analysis: {0}".format(case_id))

            for ind_id in family_parser.individuals:
                ind = family_parser.individuals[ind_id]
                logger.info("Found individual {0}".format(ind.individual_id))

                individual = Individual(
                    ind_id=ind.individual_id,
                    case_id=case_id,
                    mother=ind.mother,
                    father=ind.father,
                    sex=str(ind.sex),
                    phenotype=str(ind.phenotype),
                    variant_source=vcf,
                )
                individuals.append(individual)

        elif vcf:
            # read individuals from vcf file
            case_id = os.path.basename(vcf)
            head = HeaderParser()
            handle = get_vcf_handle(infile=vcf)
            for line in handle:
                line = line.rstrip()
                if line.startswith('#'):
                    if line.startswith('##'):
                        head.parse_meta_data(line)
                    else:
                        head.parse_header_line(line)
                else:
                    break

            for index, ind in enumerate(head.individuals):
                # If we only have a vcf file we can not get metadata about the
                # individuals
                individual = Individual(
                    ind_id=ind,
                    case_id=case_id,
                    variant_source=vcf,
                )
                individuals.append(individual)

                logger.debug("Found individual {0} in {1}".format(
                    ind, vcf))

        return individuals



import logging

from gemini import GeminiQuery

from puzzle.plugins import BaseVariantMixin

from puzzle.models import (Compound, Variant, Gene, Genotype, Transcript,)
from puzzle.utils import (get_most_severe_consequence, get_omim_number,
                          get_cytoband_coord)


logger = logging.getLogger(__name__)

class VariantMixin(BaseVariantMixin):
    """Class to store variant specific functions for gemini plugin"""


    def build_gemini_query(self, query, extra_info):
        """Append sql to a gemini query

        Args:
            query(str): The gemini query
            extra_info(str): The text that should be added

        Return:
            extended_query(str)
        """
        if 'WHERE' in query:
            return "{0} AND {1}".format(query, extra_info)
        else:
            return "{0} WHERE {1}".format(query, extra_info)

    def variants(self, case_id, skip=0, count=30, filters=None):
        """Return count variants for a case.

        This function needs to have different behaviours based on what is asked
        for. It should allways try to give minimal information back to improve
        on speed. For example, if consequences are not asked for we will not
        build all transcripts. If not sv variants we will not build sv
        coordinates.
        So the minimal case is to just show what is asked for in the variants
        interface.

            Args:
                case_id (str): A gemini db
                skip (int): Skip first variants
                count (int): The number of variants to return
                filters (dict): A dictionary with filters. Currently this will
                look like: {
                    gene_list: [] (list of hgnc ids),
                    frequency: None (float),
                    cadd: None (float),
                    consequence: [] (list of consequences),
                    impact_severities: [] (list of consequences),
                    genetic_models [] (list of genetic models)
                }

        """
        filters = filters or {}
        logger.debug("Looking for variants in {0}".format(case_id))

        limit = count + skip

        gemini_query = filters.get('gemini_query') or "SELECT * from variants"

        any_filter = False

        if filters.get('frequency'):
            frequency = filters['frequency']

            extra_info = "(max_aaf_all < {0} or max_aaf_all is"\
                         " Null)".format(frequency)
            gemini_query = self.build_gemini_query(gemini_query, extra_info)

        if filters.get('cadd'):
            cadd_score = filters['cadd']

            extra_info = "(cadd_scaled > {0})".format(cadd_score)
            gemini_query = self.build_gemini_query(gemini_query, extra_info)

        if filters.get('gene_ids'):
            gene_list = [gene_id.strip() for gene_id in filters['gene_ids']]

            gene_string = "gene in ("
            for index, gene_id in enumerate(gene_list):
                if index == 0:
                    gene_string += "'{0}'".format(gene_id)
                else:
                    gene_string += ", '{0}'".format(gene_id)
            gene_string += ")"

            gemini_query = self.build_gemini_query(gemini_query, gene_string)


        if filters.get('impact_severities'):
            severities_list = [severity.strip()
                    for severity in filters['impact_severities']]
            severity_string = "impact_severity in ("
            for index, severity in enumerate(severities_list):
                if index == 0:
                    severity_string += "'{0}'".format(severity)
                else:
                    severity_string += ", '{0}'".format(severity)
            severity_string += ")"

            gemini_query = self.build_gemini_query(gemini_query, severity_string)

        filtered_variants = self._variants(
            case_id=case_id,
            gemini_query=gemini_query,
            filters=filters,
        )

        if filters.get('consequence'):
            consequences = set(filters['consequence'])

            filtered_variants = (variant for variant in filtered_variants if
                set(variant.consequences).intersection(consequences))

        if filters.get('sv_len'):
            sv_len = int(filters['sv_len'])
            filters = (variant for variant in filtered_variants if
                variant.sv_len >= sv_len)

        for index, variant_obj in enumerate(filtered_variants):
            if index >= skip:
                if index < limit:
                    yield variant_obj
                else:
                    break

    def variant(self, case_id, variant_id):
        """Return a specific variant.

            We solve this by building a gemini query and send it to _variants

            Args:
                case_id (str): Path to a gemini database
                variant_id (int): A gemini variant id

            Returns:
                variant_obj (dict): A puzzle variant

        """
        #Use the gemini id for fast lookup
        variant_id = int(variant_id)
        gemini_query = "SELECT * from variants WHERE variant_id = {0}".format(
            variant_id
        )

        individuals = []
        # Get the individuals for the case
        case_obj = self.case(case_id)
        for individual in case_obj.individuals:
            individuals.append(individual)

        gq = GeminiQuery(self.db)
        gq.run(gemini_query)

        for gemini_variant in gq:
            variant = self._format_variant(
                gemini_variant=gemini_variant,
                individual_objs=individuals,
                index=gemini_variant['variant_id']
            )
            return variant

        return None

    def _get_genotypes(self, gemini_variant, individual_objs):
        """Add the genotypes for a variant for all individuals

            Args:
                gemini_variant (GeminiQueryRow): The gemini variant
                individual_objs (list(dict)): A list of Individuals

            Returns:
                individuals (list) A list of Genotypes
        """
        individuals = []
        for ind in individual_objs:
            index = ind.ind_index
            individuals.append(Genotype(
                sample_id=ind.ind_id,
                genotype=gemini_variant['gts'][index],
                case_id=ind.case_id,
                phenotype=ind.phenotype,
                ref_depth=gemini_variant['gt_ref_depths'][index],
                alt_depth=gemini_variant['gt_alt_depths'][index],
                depth=gemini_variant['gt_depths'][index],
                genotype_quality=gemini_variant['gt_quals'][index]
            ))

        return individuals

    def _get_hgnc_symbols(self, gemini_variant):
        """Get the hgnc symbols for all transcripts in a variant

        Right now this function only use the
            Args:
                gemini_variant (GeminiQueryRow): The gemini variant

            Returns:
                hgnc_symbols(list(str)): List hgnc symbols

        """

        # query = "SELECT * from variant_impacts WHERE variant_id = {0}".format(
        #     gemini_variant['variant_id']
        # )
        # gq = GeminiQuery(self.db)
        # gq.run(query)

        # hgnc_symbols = set([transcript['gene']])
        hgnc_symbols = [gemini_variant['gene']]

        return hgnc_symbols

    def _get_transcripts(self, gemini_variant):
        """Return a Transcript object

        Go through all transcripts found for the variant

            Args:
                gemini_variant (GeminiQueryRow): The gemini variant

            Yields:
                transcript (puzzle.models.Transcript)

        """
        query = "SELECT * from variant_impacts WHERE variant_id = {0}".format(
            gemini_variant['variant_id']
        )

        gq = GeminiQuery(self.db)
        gq.run(query)

        for genimi_transcript in gq:
            transcript = Transcript(
                hgnc_symbol=genimi_transcript['gene'],
                transcript_id=genimi_transcript['transcript'],
                consequence=genimi_transcript['impact_so'],
                biotype=genimi_transcript['biotype'],
                polyphen=genimi_transcript['polyphen_pred'],
                sift=genimi_transcript['sift_pred'],
                HGVSc=genimi_transcript['codon_change'],
                HGVSp=genimi_transcript['aa_change']
                )
            yield transcript

    def _variants(self, case_id, gemini_query, filters=None):
        """Return variants found in the gemini database

            Args:
                case_id (str): The case for which we want to see information
                gemini_query (str): What variants should be chosen
                filters (dict): A dictionary with filters

            Yields:
                variant_obj (dict): A Variant formatted dictionary
        """

        gq = GeminiQuery(self.db)

        gq.run(gemini_query)

        individuals = []
        # Get the individuals for the case
        case = self.case(case_id)
        for individual in case.individuals:
            individuals.append(individual)

        index = 0
        for gemini_variant in gq:
            # Check if variant is non ref in the individuals
            variant = None
            if self.variant_type == 'sv':
                index += 1
                variant = self._format_sv_variants(
                        gemini_variant=gemini_variant,
                        index=index,
                        filters=filters
                        )
            else:
                if self._is_variant(gemini_variant, individuals):
                    index += 1
                    logger.debug("Updating index to: {0}".format(index))

                    variant = self._format_variants(
                        gemini_variant=gemini_variant,
                        index=index,
                        filters=filters
                        )

            if variant:

                yield variant

    def _get_puzzle_variant(self, gemini_variant, index):
        """Take a gemini variant and return a basic puzzle variant

            For the overview we only need limited variant information
        """
        variant_dict = {
            'CHROM':gemini_variant['chrom'].lstrip('chrCHR'),
            'POS':str(gemini_variant['start']),
            'ID':gemini_variant['rs_ids'],
            'REF':gemini_variant['ref'],
            'ALT':gemini_variant['alt'],
            'QUAL':gemini_variant['qual'],
            'FILTER':gemini_variant['filter']
        }

        variant = Variant(**variant_dict)
        variant['index'] = index
        # Use the gemini id for fast search
        variant.update_variant_id(gemini_variant['variant_id'])

        #Add the most severe consequence
        variant['most_severe_consequence'] = gemini_variant['impact_so']

        #Add the impact severity
        variant['impact_severity'] = gemini_variant['impact_severity']

        max_freq = gemini_variant['max_aaf_all']
        if max_freq:
            variant.set_max_freq(max_freq)

        #### Check the impact annotations ####
        if gemini_variant['cadd_scaled']:
            variant['cadd_score'] = gemini_variant['cadd_scaled']

        return variant

    def _format_variants(self, gemini_variant, index=0, filters=None):
        """Format the variant for the variants view

            We want to have it's own function for doing this since it includes
            much less information than in the variant view

            Args:
                gemini_variant (GeminiQueryRow): The gemini variant
                ind_objs (list(puzzle.models.individual))
                index(int): The index of the variant

            Returns:
                variant (dict): A Variant object
        """
        variant = self._get_puzzle_variant(gemini_variant, index)

        if filters.get('consequence'):
            #If filter for consequence we need to parse the transcript
            #information
            if len(filters['consequence']) > 0:
                for transcript in self._get_transcripts(gemini_variant):
                    variant.add_transcript(transcript)

                self._add_consequences(variant)
        else:
            for hgnc_symbol in self._get_hgnc_symbols(gemini_variant):
                variant.add_transcript(Transcript(
                                hgnc_symbol=hgnc_symbol,
                                transcript_id='dummy',
                                consequence='dummy')
                                )

        for gene in self._get_genes(variant):
            variant.add_gene(gene)


        return variant

    def _format_sv_variants(self, gemini_variant, index=0, filters=None):
        """Format the variant for the sv variants view

            We want to have it's own function for doing this since it includes
            much less information than in the variant view

            Args:
                gemini_variant (GeminiQueryRow): The gemini variant
                ind_objs (list(puzzle.models.individual))
                index(int): The index of the variant

            Returns:
                variant (dict): A Variant object
        """
        variant = self._get_puzzle_variant(gemini_variant, index)

        variant.sv_type = gemini_variant['sub_type']
        variant.stop = int(gemini_variant['end'])

        ##TODO this whould be replaced with a faster lookup via phizz
        for transcript in self._get_transcripts(gemini_variant):
            variant.add_transcript(transcript)
        ##TODO same as above
        for gene in self._get_genes(variant):
            variant.add_gene(gene)

        self._add_sv_coordinates(variant)

        return variant

    def _format_variant(self, gemini_variant, individual_objs, index=0):
        """Make a puzzle variant from a gemini variant

            Args:
                gemini_variant (GeminiQueryRow): The gemini variant
                individual_objs (list(dict)): A list of Individuals
                index(int): The index of the variant

            Returns:
                variant (dict): A Variant object
        """
        variant = self._get_puzzle_variant(gemini_variant, index)

        ### GENOTYPE ANNOATTIONS ###
        #Get the genotype info
        individual_genotypes = self._get_genotypes(
            gemini_variant=gemini_variant,
            individual_objs=individual_objs
            )

        #Add the genotype info to the variant
        for individual in individual_genotypes:
            # Add the genotype calls to the variant
            variant.add_individual(individual)

        ### POSITON ANNOATTIONS ###
        variant.start = int(variant.POS)

        #Add the sv specific coordinates
        if self.variant_type == 'sv':
            variant.sv_type = gemini_variant['sub_type']
            variant.stop = int(gemini_variant['end'])
            self._add_sv_coordinates(variant)

        ### Consequence and region annotations
        #Add the transcript information
        for transcript in self._get_transcripts(gemini_variant):
            variant.add_transcript(transcript)

        #Add the genes based on the hgnc symbols
        hgnc_symbols = (transcript.hgnc_symbol for transcript in variant.transcripts)
        for gene in self._get_genes(variant):
            variant.add_gene(gene)


        # We use the prediction in text
        polyphen = gemini_variant['polyphen_pred']
        if polyphen:
            variant.add_severity('Polyphen', polyphen)

        # We use the prediction in text
        sift = gemini_variant['sift_pred']
        if sift:
            variant.add_severity('SIFT', sift)


        #### Frequencies ####
        thousand_g = gemini_variant['aaf_1kg_all']
        if thousand_g:
            variant['thousand_g'] = float(thousand_g)
            variant.add_frequency(name='1000GAF', value=float(thousand_g))

        exac = gemini_variant['aaf_exac_all']
        if exac:
            variant.add_frequency(name='EXaC', value=float(exac))

        esp = gemini_variant['aaf_esp_all']
        if esp:
            variant.add_frequency(name='ESP', value=float(esp))

        return variant

    def _is_variant(self, gemini_variant, ind_objs):
        """Check if the variants is a variation in any of the individuals

        This means that at least one of the individuals should have the
        alternative symbol.

        Args:
            gemini_variant (GeminiQueryRow): The gemini variant
            ind_objs (list(puzzle.models.individual)): A list of individuals to check

        Returns:
            bool : If any of the individuals has the variant
        """

        alt = gemini_variant['alt']
        indexes = (ind.ind_index for ind in ind_objs)
        #Merge all genotypes into one string
        genotypes = "".join([gemini_variant['gts'][index] for index in indexes])
        #Check if the alternative allele is found within the genotypes
        if alt in genotypes:
            return True

        return False



import os
import logging

from gemini import GeminiQuery

from puzzle.models import (Case, Individual)

from puzzle.plugins import BaseCaseMixin


logger = logging.getLogger(__name__)

class CaseMixin(BaseCaseMixin):
    """Class to store methods that deal with Cases in geimni plugin"""

    def cases(self, pattern=None):
        """Return all cases.
        
            Args:
                pattern: Allways None in gemini adapter
            
            Returns:
                case_objs(An iterator with Cases)
        """

        return self.case_objs

    def case(self, case_id=None):
        """Return a Case object

            If no case_id is given return one case

            Args:
                case_id (str): A case id

            Returns:
                case(Case): A Case object
        """
        cases = self.cases()
        if case_id:
            for case in cases:
                if case.case_id == case_id:
                    return case
        else:
            if cases:
                return cases[0]

        return None

    def individual(self, ind_id=None):
        """Return a individual object
        
            Args:
                ind_id (str): A individual id
            
            Returns:
                individual (puzzle.models.individual)
        """
        for ind_obj in self.individuals:
            if ind_obj.ind_id == ind_id:
                return ind_obj
        return None

    def get_individuals(self, *ind_ids):
        """Return information about individuals
        
            Args:
                ind_ids (list(str)): List of individual ids
            
            Returns:
                individuals (Iterable): Iterable with Individuals
        """
        if ind_ids:
            for ind_id in ind_ids:
                for ind in self.individuals:
                    if ind.ind_id == ind_id:
                        yield ind
        else:
            for ind in self.individuals:
                yield ind

    
    def _get_cases(self, individuals):
        """Return the cases found in the database

            Args:
                individuals(list): List of Individuals

            Returns:
                case_objs(list): List of Cases
        """
        case_objs = []
        case_ids = set()
        logger.info("Looking for cases in {0}".format(
            self.db
        ))
        for individual in individuals:
            case_ids.add(individual.case_id)

        for case_id in case_ids:
            logger.info("Found case {0}".format(case_id))
            case = Case(
                case_id=case_id,
                name=case_id,
                variant_source=self.db,
                variant_type=self.variant_type,
                variant_mode='gemini',
                )
            # Add the individuals to the correct case
            for individual in individuals:
                if individual.case_id == case_id:
                    logger.info("Adding ind {0} to case {1}".format(
                        individual.name, individual.case_id
                    ))
                    case.add_individual(individual)

            case_objs.append(case)

        return case_objs

    def _get_individuals(self):
        """Return a list with the individual objects found in db

            Returns:
                individuals (list): List of Individuals

        """
        individuals = []
        gq = GeminiQuery(self.db)
        #Dictionaru with sample to index in the gemini database
        sample_to_idx = gq.sample_to_idx

        query = "SELECT * from samples"
        gq.run(query)

        for individual in gq:
            logger.info("Found individual {0} with family id {1}".format(
                individual['name'], individual['family_id']))
            
            individuals.append(
                Individual(
                    ind_id=individual['name'],
                    case_id=individual['family_id'],
                    mother=individual['maternal_id'],
                    father=individual['paternal_id'],
                    sex=individual['sex'],
                    phenotype=individual['phenotype'],
                    ind_index=sample_to_idx.get(individual['name']),
                    variant_source=self.db,
                    bam_path=None)
            )
        return individuals



# -*- coding: utf-8 -*-
import logging

from sqlite3 import DatabaseError

from gemini import GeminiQuery

from puzzle.plugins import Plugin
from . import (CaseMixin, VariantMixin)

logger = logging.getLogger(__name__)

class GeminiPlugin(CaseMixin, VariantMixin, Plugin):
    """This is the base class for puzzle plugins

        Args:
            db(str): Path to gemini database
            vtype: Variant type (snv or sv)

    """

    def __init__(self, db=None, vtype='snv'):
        super(GeminiPlugin, self).__init__()
        logger.debug("Setting self.db to {0}".format(db))
        self.db = db
        logger.debug("Setting variant type to {0}".format(vtype))
        self.variant_type = vtype

        if self.db:
            logger.info("Check if database is in correct format")
            self.test_gemini_db()
            
            self.individuals = self._get_individuals()
            self.case_objs = self._get_cases(self.individuals)

        self.filters.can_filter_gene = True
        self.filters.can_filter_frequency = True
        self.filters.can_filter_cadd = True
        self.filters.can_filter_consequence = True
        self.filters.can_filter_impact_severity = True
        self.filters.can_query_gemini=True

    def test_gemini_db(self):
        """Check if self.db is a valid gemini database
            
            Raises sqlite3.DatabaseError if not a valid databse
        """
        gq = GeminiQuery(self.db)
        return True

    def init_app(self, app):
        """Initialize plugin via Flask."""
        pass



# -*- coding: utf-8 -*-
from flask import (abort, current_app as app, Blueprint, render_template,
                   request)

from puzzle.constants import (INHERITANCE_MODELS_SHORT, SO_TERMS, SV_TYPES,
                              IMPACT_LEVELS)

BP_NAME = __name__.split('.')[-2]
blueprint = Blueprint(BP_NAME, __name__, url_prefix='/variants',
                      template_folder='templates', static_folder='static',
                      static_url_path="/{}/static".format(BP_NAME))


@blueprint.route('/<case_id>')
def variants(case_id):
    """Show all variants for a case."""
    filters = parse_filters()
    is_active = any(value for value in filters.values()
                    if not isinstance(value, dict))
    variants = app.db.variants(
        case_id,
        skip=filters['skip'],
        filters={
            'gene_ids': filters['gene_symbols'],
            'frequency': filters.get('frequency'),
            'cadd': filters.get('cadd'),
            'sv_len': filters.get('sv_len'),
            'consequence': filters['selected_consequences'],
            'genetic_models': filters['selected_models'],
            'sv_types': filters['selected_sv_types'],
            'gene_lists': filters['gene_lists'],
            'impact_severities': filters['impact_severities'],
            'gemini_query': filters['gemini_query'],
        }
    )
    gene_lists = ([gene_list.list_id for gene_list in app.db.gene_lists()]
                  if app.config['STORE_ENABLED'] else [])
    kwargs = dict(variants=variants, case_id=case_id, db=app.db,
                  filters=filters, consequences=SO_TERMS,
                  inheritance_models=INHERITANCE_MODELS_SHORT,
                  gene_lists=gene_lists, impact_severities=IMPACT_LEVELS,
                  is_active=is_active)

    if app.db.variant_type == 'sv':
        return render_template('sv_variants.html', sv_types=SV_TYPES, **kwargs)
    else:
        return render_template('variants.html', **kwargs)


@blueprint.route('/<case_id>/<variant_id>')
def variant(case_id, variant_id):
    """Show a single variant."""
    variant = app.db.variant(case_id, variant_id)
    if variant is None:
        return abort(404, "variant not found")

    # sort compounds by score
    sorted_compounds = sorted(variant['compounds'],
                              key=lambda compound: compound['combined_score'])

    template = 'sv_variant.html' if app.db.variant_type == 'sv' else 'variant.html'
    return render_template(template, variant=variant,
                           compounds=sorted_compounds, case_id=case_id)


def parse_filters():
    """Parse variant filters from the request object."""
    genes_str = request.args.get('gene_symbol')
    filters = {}
    for key in ('frequency', 'cadd', 'sv_len'):
        try:
            filters[key] = float(request.args.get(key))
        except (ValueError, TypeError):
            pass

    filters['gene_symbols'] = genes_str.split(',') if genes_str else None
    filters['selected_models'] = request.args.getlist('inheritance_models')
    filters['selected_consequences'] = request.args.getlist('consequences')
    filters['selected_sv_types'] = request.args.getlist('sv_types')
    filters['skip'] = int(request.args.get('skip', 0))
    filters['gene_lists'] = request.args.getlist('gene_lists')
    filters['gemini_query'] = request.args.get('gemini_query')
    filters['impact_severities'] = request.args.getlist('impact_severities')

    filters['query_dict'] = {key: request.args.getlist(key) for key
                             in request.args.keys()}
    filters['query_dict'].update({'skip': (filters['skip'] + 30)})

    return filters



from puzzle.plugins import GeminiPlugin

def test_get_individuals(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    
    ind_ids = [ind.ind_id for ind in adapter.get_individuals()]
    assert set(ind_ids) == set(['NA12877', 'NA12878', 'NA12882'])

def test_get_individuals_one_ind(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    
    ind_ids = [ind.ind_id for ind in adapter.get_individuals('NA12877')]
    assert set(ind_ids) == set(['NA12877'])

def test_get_individuals_two_inds(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    
    ind_ids = [ind.ind_id for ind in adapter.get_individuals('NA12877', 'NA12878')]
    assert set(ind_ids) == set(['NA12877', 'NA12878'])

def test__get_individuals(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    
    ind_ids = [ind.ind_id for ind in adapter._get_individuals()]
    assert set(ind_ids) == set(['NA12877', 'NA12878', 'NA12882'])

def test__get_cases(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    
    individuals = adapter._get_individuals()
    case_ids = [case.case_id for case in adapter._get_cases(individuals)]
    assert set(case_ids) == set(['643594'])

def test_cases(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    
    case_ids = [case.case_id for case in adapter.cases()]
    assert set(case_ids) == set(['643594'])

def test_case(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    
    case_id = '643594'
    assert adapter.case(case_id).case_id == case_id

def test_case_no_id(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    
    case_id = '643594'
    assert adapter.case().case_id == case_id

def test_case_wrong_id(gemini_path):
    adapter = GeminiPlugin(db=gemini_path)
    
    case_id = 'hello'
    assert adapter.case(case_id) == None



# -*- coding: utf-8 -*-
from puzzle.plugins import VcfPlugin

# TODO: define a vcf in the test
vcf = "tests/fixtures/hapmap.vcf"
family_file = "tests/fixtures/hapmap.ped"

individuals = [
    "636808	ADM1059A1	0	0	1	1",
    "636808	ADM1059A2	ADM1059A1	ADM1059A3	1	2",
    "636808	ADM1059A3	0	0	2	1"
]


class MockApp(object):
    """Mock a flask app"""
    def __init__(self, db=None, family=None, family_type='ped',
                 pattern='*.vcf', mode='snv'):
        super(MockApp, self).__init__()
        self.config = {'FAMILY_FILE': family, 'FAMILY_TYPE': family_type}


def test_variants():
    vcf_plugin = VcfPlugin(root_path=vcf)
    variants = vcf_plugin.variants("hapmap.vcf")
    variant = next(variants)
    assert variant['CHROM'] == 'X'
    assert int(variant['POS']) == 84563218
    assert variant['index'] == 1

    variant = next(variants)
    assert variant['CHROM'] == '2'
    assert int(variant['POS']) == 233349186
    assert variant['index'] == 2

def test_variants_case_no_ped():
    vcf_plugin = VcfPlugin(root_path=vcf)
    variants = vcf_plugin.variants(case_id='hapmap.vcf')
    variant = next(variants)
    assert variant['CHROM'] == 'X'
    assert int(variant['POS']) == 84563218
    assert variant['index'] == 1

    variant = next(variants)
    assert variant['CHROM'] == '2'
    assert int(variant['POS']) == 233349186
    assert variant['index'] == 2

def test_variants_case_with_ped():
    vcf_plugin = VcfPlugin(root_path=vcf, case_lines=individuals, case_type='ped')
    variants = vcf_plugin.variants(case_id='636808')
    variant = next(variants)
    assert variant['CHROM'] == 'X'
    assert int(variant['POS']) == 84563218
    assert variant['index'] == 1

    variant = next(variants)
    assert variant['CHROM'] == '2'
    assert int(variant['POS']) == 233349186
    assert variant['index'] == 2


def test_variant():
    vcf_plugin = VcfPlugin(root_path=vcf)
    variant = vcf_plugin.variant('hapmap.vcf', 'X_155239821_G_A')
    assert variant['CHROM'] == 'X'
    assert variant['POS'] == '155239821'

    # get 10th variant
    variant = vcf_plugin.variant('hapmap.vcf', '3_124998098_C_A')
    assert variant['index'] == 10

def test_ped_info():
    adapter=VcfPlugin(root_path=vcf, case_lines=individuals, case_type='ped')
    assert len(adapter.individuals) == 3
    case_obj = adapter.case_objs[0]
    assert case_obj.name == "636808"

def test_vcf_case():
    adapter=VcfPlugin(root_path=vcf)
    assert len(adapter.individuals) == 3
    case_obj = adapter.case_objs[0]
    assert case_obj['name'] == 'hapmap.vcf'

def test_vcf_case_dir():
    adapter=VcfPlugin(root_path="tests/fixtures/")
    assert len(adapter.case_objs) == 3




# -*- coding: utf-8 -*-
import os
import logging

from path import path

from puzzle.plugins import Plugin
from puzzle.models import DotDict
from puzzle.utils import get_case
from . import VariantMixin, CaseMixin

logger = logging.getLogger(__name__)


class VcfPlugin(VariantMixin, CaseMixin, Plugin):
    """docstring for Plugin"""

    def __init__(self, root_path=None, case_lines=None,
                 case_type=None, pattern='*.vcf', vtype='snv',
                 case_obj=None):
        """Initialize a vcf adapter.

            When instansiating all cases are found.

            Args:
                root_path(str) : Path to a vcf file or a dir with vcf files
                case_lines(Iterable) : Lines with ped like information
                case_type(str) : Format of pedigreeinformation
                patter(str) : What pattern to search for in directory
                vtype(str) : 'snv' or 'sv'
                case_obj(puzzle.models.case) : If initialized with a case
        """
        super(VcfPlugin, self).__init__()

        self.individuals = []
        self.case_objs = []
        logger.info("Updating root path to {0}".format(root_path))
        self.root_path = root_path

        self.check_setup(case_lines)

        self.variant_type = vtype
        logger.info("Setting variant type to {0}".format(vtype))

        self.pattern = pattern
        logger.debug("Updating pattern to {0}".format(pattern))

        if root_path:
            if os.path.isdir(root_path):
                logger.info("Looking for vcf files in {0}".format(root_path))
                for vcf_file in self._find_vcfs(pattern=pattern):
                    logger.info("Found vcf {0}".format(vcf_file))
                    case_obj = get_case(variant_source=vcf_file)
                    self.case_objs.append(case_obj)
            else:
                self.case_objs.append(get_case(
                    variant_source=self.root_path,
                    case_lines=case_lines,
                    case_type=case_type
                    )
                )

            for case_obj in self.case_objs:
                for ind in case_obj.individuals:
                    self.individuals.append(ind)

        self.filters = DotDict(
            can_filter_frequency=True,
            can_filter_cadd=True,
            can_filter_consequence=True,
            can_filter_gene=True,
            can_filter_inheritance=True,
            can_filter_sv=True,
            can_filter_impact_severity=True
            
        )

    def check_setup(self, case_lines):
        """Make some small tests to see if setup is correct"""
        valid_vcf_suffixes = ('.vcf', '.vcf.gz')
        if self.root_path:
            if case_lines:
                # If family file we only allow one vcf file as input
                if not os.path.isfile(self.root_path):
                    raise SyntaxError("Variant source has to be a vcf file when"\
                                     " running with family file")
            if os.path.isfile(self.root_path):
                if not self.root_path.endswith(valid_vcf_suffixes):
                    raise SyntaxError("Vcf file has to end with with .vcf or .vcf.gz")
        return


    def _find_vcfs(self, pattern='*.vcf'):
        """Walk subdirectories and return VCF files.

            Args:
                pattern (str): What pattern to serch for in filenames

            Return:
                paths (Iterable): The paths found
        """
        return path(self.root_path).walkfiles(pattern)

    def init_app(self, app):
        """Initialize plugin via Flask."""
        pass

